/*
  Copyright(c) 2009 NEC Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called COPYING.
*/

/* Character-device driver for updating the system bios. */

#include <linux/version.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
#include <linux/device.h>
#else
#include <linux/config.h>
#endif

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>
#ifdef __x86_64__
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
#include <asm/compat.h>
#else
#include <asm/ioctl32.h>
#endif
#endif

#include "OnlineBIOSUpdate.h"

static int ReadIo (UAOpCode, void *);
static int ReadMemory (UAOpCode, void *);
static int ReadPci (UAOpCode, unsigned int *);
static int WriteIo (UAOpCode);
static int WriteMemory (UAOpCode);
static int WritePci (UAOpCode);
static int SetSpiFlashInfo (unsigned long, struct _LOCAL_DEVICE_INFO *);
static int EraseFlashSector (unsigned long, struct _LOCAL_DEVICE_INFO *);
static int EraseSpiFlashSector (unsigned long, struct _LOCAL_DEVICE_INFO *);
static int WriteFlashSector (unsigned long, struct _LOCAL_DEVICE_INFO *);
static int WriteSpiFlashSector (unsigned long, struct _LOCAL_DEVICE_INFO *);
static int GetSpiJedecId(unsigned long, struct _LOCAL_DEVICE_INFO *);
static int IssueSpiCommand (PLOCAL_DEVICE_INFO, unsigned int, unsigned int,
             unsigned int, unsigned int, BYTE **);
static int OBU_open (struct inode *, struct file *);
static int OBU_release (struct inode *, struct file *);
static ssize_t OBU_read (struct file *, char *, size_t, loff_t *);
static loff_t OBU_lseek (struct file *, loff_t, int);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
static int OBU_ioctl (struct inode *, struct file *, unsigned int, unsigned long);
#endif
#if defined(__x86_64__) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
extern int register_ioctl32_conversion (unsigned int cmd,
               int (*handler) (unsigned int,
                     unsigned int,
                     unsigned long,
                     struct file *));
extern int unregister_ioctl32_conversion (unsigned int cmd);
#endif
#if defined(CONFIG_COMPAT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
static long OBU_compat_ioctl (struct file *, unsigned int, unsigned long);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void msleep (unsigned int);
#endif

MODULE_AUTHOR ("NEC Corporation");
MODULE_LICENSE ("GPL");
MODULE_DESCRIPTION
  ("Character-device driver for updating the system bios, Version 2.0.1");

#define MAX_KMALLOC_SIZE 0x20000
// bios flash access control
#define MAX_DEVICE 1
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
static DEFINE_SPINLOCK(obu_spin);
#else
static spinlock_t obu_spin = SPIN_LOCK_UNLOCKED;
#endif
static int obu_stat = OBU_CLOSE;
static int obu_major = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
static struct class *obu_class;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void
msleep (unsigned int msecs)
{
  unsigned long timeout = msecs_to_jiffies (msecs) + 1;

  while (timeout){
    set_current_state (TASK_UNINTERRUPTIBLE);
    timeout = schedule_timeout (timeout);
  }
}
#endif

/*
 * Read io port
 */
static int ReadIo (UAOpCode OpCode, void *Buffer)
{
  int error = 0;
  OBU_DEBUG ("OBU: IO READ: addr = %x\n", OpCode.Addr);

  switch (OpCode.CmdSize)
  {
  case COMMAND_SIZE_BYTE:
    *(unsigned char *) Buffer = inb (OpCode.Addr);
    break;
  case COMMAND_SIZE_WORD:
    if (OpCode.Addr & 1){        //bad alignment
      error = -1;
    } else {
      *(unsigned short *) Buffer = inw (OpCode.Addr);
    }
    break;
  case COMMAND_SIZE_DWORD:
    if (OpCode.Addr & 3){        //bad alignment
      error = -1;
    } else {
      *(unsigned int *) Buffer = inl (OpCode.Addr);
    }
    break;
  default:
    printk ("<3>OnlineBIOSUpdate: IO READ SIZE ERROR\n");
    return -EFAULT;
  }
  OBU_DEBUG ("OBU: IO READ: data = %lx\n", *(unsigned long *) Buffer);

  return error;
}

/*
 * Read memory
 */
static int ReadMemory (UAOpCode OpCode, void *Buffer)
{
  int error = 0;
  void *virt_addr_page = NULL;
  unsigned int io_page, io_offset;

  OBU_DEBUG ("OBU: MEMORY READ: addr = %x\n", OpCode.Addr);

#ifdef __x86_64__
  io_page = OpCode.Addr & 0xfffffffffffff000;
#else
  io_page = OpCode.Addr & 0xfffff000;
#endif
  io_offset = OpCode.Addr & 0xfff;

  if ((virt_addr_page = ioremap_nocache (io_page, 0x1000)) == NULL){
    printk (KERN_ERR "OBU: can't remap memory @ 0x%x\n", io_page);
    return -EIO;
  }
  switch (OpCode.CmdSize)
  {
  case COMMAND_SIZE_BYTE:
    *(unsigned char *) Buffer = readb (virt_addr_page + io_offset);
    break;
  case COMMAND_SIZE_WORD:
    if (io_offset > 0xffe){        //bad alignment
      printk (KERN_ERR "OBU: memory access bad alignment @ 0x%x\n", OpCode.Addr);
      error = -EIO;
    } else {
      *(unsigned short *) Buffer = readw (virt_addr_page + io_offset);
    }
    break;
  case COMMAND_SIZE_DWORD:
    if (io_offset > 0xffc){        //bad alignment
      printk (KERN_ERR "OBU: memory access bad alignment @ 0x%x\n", OpCode.Addr);
      error = -EIO;
    } else {
      *(unsigned int *) Buffer = readl (virt_addr_page + io_offset);
    }
    break;
  default:
    printk ("<3>OnlineBIOSUpdate: MEM READ SIZE ERROR\n");
    error = -EFAULT;
  }
  OBU_DEBUG ("OBU: phy_addr=%lx -> vir_addr_page=%lx\n",
        (unsigned long) OpCode.Addr, (unsigned long) virt_addr_page);

  iounmap (virt_addr_page);
  return error;
}

/*
 * Read pci configuration space
 */
static int ReadPci (UAOpCode OpCode, unsigned int *Buffer)
{
  int error = 0;
  int offset;
  struct pci_dev *pcidev;

  OBU_DEBUG ("OBU: PCI READ: addr = %x\n", OpCode.Addr);

  offset = GET_REGISTER_NUMBER (OpCode.Addr);
  pcidev = pci_find_slot ((unsigned int) GET_BUS_NUMBER (OpCode.Addr),
           (unsigned int) GET_DEVFUN_NUMBER (OpCode.Addr));
  if (pcidev == NULL){
    printk ("<3>OnlineBIOSUpdate: ERROR PCI DEVICE NOT FOUND\n");
    return -ENODEV;
  }
  switch (OpCode.CmdSize)
  {
  case COMMAND_SIZE_BYTE:
    error = pci_read_config_byte (pcidev, offset, (u8 *) Buffer);
    break;
  case COMMAND_SIZE_WORD:
    error = pci_read_config_word (pcidev, offset, (u16 *) Buffer);
    break;
  case COMMAND_SIZE_DWORD:
    error = pci_read_config_dword (pcidev, offset, (u32 *) Buffer);
    break;
  default:
    printk ("<3>OnlineBIOSUpdate: PCI READ SIZE ERROR\n");
    return -EFAULT;
  }
  OBU_DEBUG ("OBU: PCI READ: data = %lx\n", (unsigned long) Buffer);

  return error;
}

/*
 * Write io port
 */
static int WriteIo (UAOpCode OpCode)
{
  int error = 0;

  OBU_DEBUG ("OBU: IO WRITE address:%x = data:%x\n", OpCode.Addr,
        OpCode.Data);

  switch (OpCode.CmdSize)
  {
  case COMMAND_SIZE_BYTE:
    outb (OpCode.Data, OpCode.Addr);
    break;
  case COMMAND_SIZE_WORD:
    if (OpCode.Addr & 1){        //bad alignment
      error = -1;
    } else {
     outw (OpCode.Data, OpCode.Addr);
    }
    break;
  case COMMAND_SIZE_DWORD:
    if (OpCode.Addr & 3){        //bad alignment
      error = -1;
    } else {
      outl (OpCode.Data, OpCode.Addr);
    }
    break;
  default:
    printk ("<3>OnlineBIOSUpdate: IO WRITE SIZE ERROR\n");
    return -EFAULT;
  }
  return error;
}

/*
 * Write memory
 */
static int WriteMemory (UAOpCode OpCode)
{
  int error = 0;
  void *virt_addr_page = NULL;
  unsigned int io_page, io_offset;

  OBU_DEBUG ("OBU: MEM WRITE address:%x = data:%x\n", OpCode.Addr,
        OpCode.Data);

#ifdef __x86_64__
  io_page = OpCode.Addr & 0xfffffffffffff000;
#else
  io_page = OpCode.Addr & 0xfffff000;
#endif
  io_offset = OpCode.Addr & 0xfff;

  if ((virt_addr_page = ioremap_nocache (io_page, 0x1000)) == NULL){
    printk (KERN_ERR "OBU: can't remap memory @ 0x%x\n", io_page);
    return -EIO;
  }
  switch (OpCode.CmdSize)
  {
  case COMMAND_SIZE_BYTE:
    *((unsigned char *) (virt_addr_page + io_offset)) = (unsigned char) OpCode.Data;
    break;
  case COMMAND_SIZE_WORD:
    if (io_offset > 0xffe){        //bad alignment
      printk (KERN_ERR "OBU: memory access bad alignment @ 0x%x\n", OpCode.Addr);
      error = -EIO;
    } else {
      *((unsigned short *) (virt_addr_page + io_offset)) = (unsigned short) OpCode.Data;
    }
    break;
  case COMMAND_SIZE_DWORD:
    if (io_offset > 0xffc){        //bad alignment
      printk (KERN_ERR "OBU: memory access bad alignment @ 0x%x\n", OpCode.Addr);
      error = -EIO;
    } else {
      *((unsigned int *) (virt_addr_page + io_offset)) = (unsigned int) OpCode.Data;
    }
    break;
  default:
    printk ("<3>OnlineBIOSUpdate: MEM WRITE SIZE ERROR\n");
    error = -EFAULT;
  }
  iounmap (virt_addr_page);
  return error;
}

/*
 * Write pci configuration space
 */
static int WritePci (UAOpCode OpCode)
{
  int error = 0;
  int offset;
  struct pci_dev *pcidev;

  OBU_DEBUG ("OBU: PCI WRITE address:%x = data:%x\n", OpCode.Addr,
        OpCode.Data);

  offset = GET_REGISTER_NUMBER (OpCode.Addr);
  pcidev = pci_find_slot ((unsigned int) GET_BUS_NUMBER (OpCode.Addr),
           (unsigned int) GET_DEVFUN_NUMBER (OpCode.Addr));
  if (pcidev == NULL){
    printk ("<3>OnlineBIOSUpdate: ERROR PCI DEVICE NOT FOUND\n");
    return -ENODEV;
  }
  switch (OpCode.CmdSize)
  {
  case COMMAND_SIZE_BYTE:
    error = pci_write_config_byte (pcidev, offset, (u8) OpCode.Data);
    break;
  case COMMAND_SIZE_WORD:
    if (OpCode.Addr & 1){        //bad alignment
      error = -1;
    } else {
      error = pci_write_config_word (pcidev, offset, (u16) OpCode.Data);
    }
    break;
  case COMMAND_SIZE_DWORD:
    if (OpCode.Addr & 3){        //bad alignment
      error = -1;
    } else {
      error = pci_write_config_dword (pcidev, offset, (u8) OpCode.Data);
    }
    break;
  default:
    printk ("<3>OnlineBIOSUpdate: PCI WRITE SIZE ERROR\n");
    return -EFAULT;
  }
  return error;
}

/*
 * Set SPI Flash Info
 */
static int SetSpiFlashInfo (unsigned long arg, struct _LOCAL_DEVICE_INFO *pLDI)
/*++

Routine Description:
    This routine processes the IOCTLs which set the Flash Map Information.

Arguments:
    
    pLDI        - our local device data
Return Value:
    STATUS_SUCCESS           -- OK

    -EINVAL -- Invalid argument

--*/
{
  unsigned int IchSpiBar;
  int i;
  UAOpCode OpCode;


  if (copy_from_user(&pLDI->IchSpi, (void *) arg, sizeof (SPI_ICH_FLASH_INFO))){
      return -EFAULT;
  } else {
    if (pLDI->IchSpi.Signature != SPI_ICH8_SIGNATURE){
      return -EINVAL;
    }

  }

  OpCode.Addr = pLDI->IchSpi.IchPciRcrbAddress;
  OpCode.CmdSize = sizeof (IchSpiBar);
  ReadPci (OpCode, &IchSpiBar);
  IchSpiBar &= pLDI->IchSpi.IchPciRcrbMask;
  IchSpiBar += pLDI->IchSpi.IchSpiBarOffset;

  // Map Device Physical Memory to OS Logical Memory
  if ((pLDI->pMappedSpiRegs = ioremap_nocache (IchSpiBar, pLDI->IchSpi.IchSpiMappedSize)) == 0){
      printk (KERN_ERR "OnlineBIOSUpdate: Unable to remap memory.\n");
      return -ENODEV;
  }

#ifdef DEBUG
  printk ("<7>OBU: ICH SPI Addr:%x\n", IchSpiBar);
  printk ("<7>OBU: ICH SPI Size:%d\n", pLDI->IchSpi.IchSpiMappedSize);
#endif

  //
  // Set spi software method registers.
  //
  SPI_ICH_WRITE_REG ((WORD *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiTypeReg), pLDI->IchSpi.IchSpiType);
  if (CHECK_WRITE_DATA((WORD *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiTypeReg), pLDI->IchSpi.IchSpiType) == FALSE){
    iounmap (pLDI->pMappedSpiRegs);
    return -EINVAL;
  }
  SPI_ICH_WRITE_REG ((BYTE *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiPrefixReg0), pLDI->IchSpi.IchSpiPrefix0);
  if (CHECK_WRITE_DATA((BYTE *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiPrefixReg0), pLDI->IchSpi.IchSpiPrefix0) == FALSE){
    iounmap (pLDI->pMappedSpiRegs);
    return -EINVAL;
  }
  SPI_ICH_WRITE_REG ((BYTE *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiPrefixReg0) + 1, pLDI->IchSpi.IchSpiPrefix1);
  if (CHECK_WRITE_DATA((BYTE *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiPrefixReg0) + 1, pLDI->IchSpi.IchSpiPrefix1) == FALSE){
    iounmap (pLDI->pMappedSpiRegs);
    return -EINVAL;
  }
  for (i = 0; i < SPI_ICH_NUM_CMD_REGS; i++){
    SPI_ICH_WRITE_REG ((BYTE *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiCmdReg0) + i, pLDI->IchSpi.IchCmd[i]);
    if (CHECK_WRITE_DATA((BYTE *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiCmdReg0) + i, pLDI->IchSpi.IchCmd[i]) == FALSE){
      iounmap (pLDI->pMappedSpiRegs);
      return -EINVAL;
    }
  }

  pLDI->SpiUsed = TRUE;

  return STATUS_SUCCESS;
}

/*
 * Erase Flash Sector
 */
static int EraseFlashSector (unsigned long arg, struct _LOCAL_DEVICE_INFO *pLDI)
/*++

Routine Description:
    This routine processes the IOCTLs which erase the Flash Block.

Arguments:
    
    pLDI        - our local device data

Return Value:
    STATUS_SUCCESS           -- OK

    -EINVAL -- Invalid argument

--*/
{
  unsigned long i;
  unsigned long TimeOut;
  unsigned long LoopCount;


  //unsigned char *FlashReg8;
  //unsigned short *FlashReg16;
  unsigned int *FlashReg32;
  unsigned int SectorBase;
  unsigned int SectorSize;
  unsigned int NumOps;
  UAOpCode *OpCode;
  unsigned int Offset;
  unsigned int Period;
  unsigned int Flags;
  unsigned char *StatReg = NULL;

  if (get_user (SectorBase, ((unsigned int *) arg)) != 0)
    return -EFAULT;
  if (get_user (SectorSize, ((unsigned int *) arg + 1)) != 0)
    return -EFAULT;
  if (get_user (NumOps, ((unsigned int *) arg + 2)) != 0)
    return -EFAULT;
  if (get_user (Flags, ((unsigned int *) arg + 3)) != 0)
    return -EFAULT;
  if (get_user (Period, ((unsigned int *) arg + 4)) != 0)
    return -EFAULT;

  OpCode = (UAOpCode *) kmalloc (sizeof (UAOpCode) * NumOps, GFP_KERNEL);
  if (copy_from_user (OpCode, (unsigned int *) arg + 5, sizeof (UAOpCode) * NumOps))
  {
    kfree (OpCode);
    return -EFAULT;
  }

  OBU_DEBUG ("OBU: SectorBase: %x, SectorSize:%x, NumOps:%x\n", SectorBase, SectorSize, NumOps);
#ifdef DEBUG
  for (i = 0; i < NumOps; i++){
    printk ("<6>OBU: cmd:%x, size:%x, addr:%x, data:%x\n",
       OpCode[i].Cmd, OpCode[i].CmdSize, OpCode[i].Addr, OpCode[i].Data);
  }
#endif

  for (i = 0; i < NumOps; i++){
    switch ((unsigned char) OpCode[i].CmdSize){
    case COMMAND_SIZE_BYTE:
      writeb ((unsigned char) OpCode[i].Data, (pLDI->pMappedFlashAddress + SectorBase + OpCode[i].Addr));
      break;
    case COMMAND_SIZE_WORD:
      writew ((unsigned short) OpCode[i].Data, (pLDI->pMappedFlashAddress + SectorBase + OpCode[i].Addr));
      break;
    case COMMAND_SIZE_DWORD:
      writel ((unsigned int) OpCode[i].Data, (pLDI->pMappedFlashAddress + SectorBase + OpCode[i].Addr));
      break;
    default:
      printk ("<3>OnlineBIOSUpdate: ERASE FLASH SECTOR SIZE ERROR\n");
      kfree (OpCode);
      return -EINVAL;
    }
  }

  //
  // Default timeout period is 1s.
  // Max erase time may 15s
  //
  kfree (OpCode);
  TimeOut = 1500;    // 10 msec * 1500 = 15 sec
  LoopCount = 0;

  //
  // Support STMicro FWH
  //
  if (Flags & FLAGS_STMICRO_FWH){
    while (LoopCount < TimeOut){  // to avoid dead loop
      msleep (10);
      StatReg = (unsigned char *) (pLDI->pMappedFlashAddress + SectorBase);
      if ((*StatReg & 0x80) && !(*StatReg & 0x3A)){
        break;
      }
      LoopCount++;
    }
    if (LoopCount == TimeOut){
      OBU_DEBUG ("OBU: ERASE FLASH SECTOR ERROR\n");
      return -EFAULT;
    }
    *StatReg = 0xff;
    udelay (1);
  } else {
    while (LoopCount < TimeOut){  // to avoid dead loop
      msleep (10);
      for (Offset = 0; Offset < SectorSize; Offset += sizeof (unsigned int)){
        FlashReg32 = (unsigned int *) (pLDI->pMappedFlashAddress + SectorBase + Offset);
        if (*FlashReg32 != 0xffffffff){
#ifdef DEBUG
          printk ("<7>OBU: DON'T ERASE SECTOR\n");
          printk ("<7>OBU: %ld times Offset:%x, Data:%x, MapAddr:%p\n", LoopCount + 1, Offset, *FlashReg32, FlashReg32);
#endif
          break;
        }
      }
      if (Offset == SectorSize){
        break;
      }
      LoopCount++;
    }
    if (LoopCount == TimeOut){
      OBU_DEBUG ("OBU: ERASE FLASH SECTOR ERROR\n");
      return -EFAULT;
    }
  }

  return STATUS_SUCCESS;
}

/*
 * Write Flash Sector
 */
static int WriteFlashSector (unsigned long arg, struct _LOCAL_DEVICE_INFO *pLDI)
/*++

Routine Description:
    This routine processes the IOCTLs which write the Flash Block.

Arguments:
    
    pLDI        - our local device data

Return Value:
    STATUS_SUCCESS           -- OK

    -EINVAL -- Invalid argument

--*/
{
  unsigned long i;
  unsigned int *position;
  unsigned long TimeOut;
  unsigned long LoopCount;

  unsigned char *FlashReg8 = NULL;
  unsigned short *FlashReg16 = NULL;
  unsigned int *FlashReg32 = NULL;
  unsigned int SectorBase;
  unsigned int SectorSize;
  unsigned int NumOps;
  UAOpCode *OpCode;
  void *WriteData;
  unsigned int Offset = 0;
  unsigned char DataSize = 0;
  int IsWriteComplete = 0;
  unsigned int Flags;
  unsigned char *StatReg = NULL;

  if (get_user (SectorBase, ((unsigned int *) arg)) != 0)
    return -EFAULT;
  if (get_user (SectorSize, ((unsigned int *) arg + 1)) != 0)
    return -EFAULT;
  if (get_user (NumOps, ((unsigned int *) arg + 2)) != 0)
    return -EFAULT;
  if (get_user (Flags, ((unsigned int *) arg + 3)) != 0)
    return -EFAULT;

  OpCode = (UAOpCode *) kmalloc (sizeof (UAOpCode) * NumOps, GFP_KERNEL);
  if (copy_from_user(OpCode, (unsigned int *) arg + 4, sizeof (UAOpCode) * NumOps)){
    kfree (OpCode);
    return -EFAULT;
  }

  position = (unsigned int *) arg + 4 + (sizeof (UAOpCode) * NumOps / sizeof (int));
#ifdef DEBUG
  printk ("<7>OBU: SectorBase:%x, SectorSize:%x, NumOps:%x\n", SectorBase, SectorSize, NumOps);
  for (i = 0; i < NumOps; i++){
    printk ("<7>OBU: cmd:%x, size:%x, addr:%x, data:%x\n",
       OpCode[i].Cmd, OpCode[i].CmdSize, OpCode[i].Addr,
       OpCode[i].Data);
  }
#endif

  WriteData = kmalloc (SectorSize, GFP_KERNEL);
  if (copy_from_user (WriteData, (const void *) position, SectorSize)){
    kfree (OpCode);
    kfree (WriteData);
    return -EFAULT;
  }

  DataSize = (unsigned char) OpCode[NumOps - 1].CmdSize;

  for (Offset = 0; Offset < SectorSize; Offset += DataSize){
    switch (DataSize){
    case COMMAND_SIZE_BYTE:
      if (((unsigned char *) WriteData)[Offset] == 0xff){
        continue;
      }
      break;
    case COMMAND_SIZE_WORD:
      if (((unsigned short *) WriteData)[Offset / DataSize] == 0xffff){
        continue;
      }
      break;
    case COMMAND_SIZE_DWORD:
      if (((unsigned int *) WriteData)[Offset / DataSize] == 0xffffffff){
        continue;
      }
      break;
    default:
      printk("<3>OnlineBIOSUpdate: WRITE FLASH SECTOR DATA SIZE ERROR\n");
      kfree (OpCode);
      kfree (WriteData);
      return -EINVAL;
    }
    for (i = 0; i < NumOps - 1; i++){
      switch ((unsigned char) OpCode[i].CmdSize){
      case COMMAND_SIZE_BYTE:
        writeb ((unsigned char) OpCode[i].Data, (pLDI->pMappedFlashAddress + SectorBase + OpCode[i].Addr));
        break;
      case COMMAND_SIZE_WORD:
        writew ((unsigned short) OpCode[i].Data, (pLDI->pMappedFlashAddress + SectorBase + OpCode[i].Addr));
        break;
      case COMMAND_SIZE_DWORD:
        writel ((unsigned int) OpCode[i].Data, (pLDI->pMappedFlashAddress + SectorBase + OpCode[i].Addr));
        break;
      default:
        printk("<3>OnlineBIOSUpdate: CANNOT EXECUTE FLASH WRITE COMMAND\n");
        kfree (OpCode);
        kfree (WriteData);
        return -EINVAL;
      }
    }

    // Write data to write address
    switch (DataSize){
    case COMMAND_SIZE_BYTE:
      FlashReg8 = (unsigned char *) (pLDI->pMappedFlashAddress + SectorBase + Offset);
      *FlashReg8 = ((unsigned char *) WriteData)[Offset];
      break;
    case COMMAND_SIZE_WORD:
      FlashReg16 = (unsigned short *) (pLDI->pMappedFlashAddress + SectorBase + Offset);
      *FlashReg16 = ((unsigned short *) WriteData)[Offset / DataSize];
      break;
    case COMMAND_SIZE_DWORD:
      FlashReg32 = (unsigned int *) (pLDI->pMappedFlashAddress + SectorBase + Offset);
      *FlashReg32 = ((unsigned int *) WriteData)[Offset / DataSize];
      break;
    default:
      printk ("<3>OnlineBIOSUpdate: WRITE FLASH SECTOR SIZE ERROR\n");
      kfree (OpCode);
      kfree (WriteData);
      return -EINVAL;
    }

    IsWriteComplete = 0;
    TimeOut = 100;
    LoopCount = 0;

    if (Flags & FLAGS_STMICRO_FWH){
      while (LoopCount < TimeOut){   // to avoid dead loop
        udelay (10);
        StatReg = (unsigned char *) (pLDI->pMappedFlashAddress + SectorBase);
        if ((*StatReg & 0x80) && !(*StatReg & 0x1A)){
          break;
        }
        LoopCount++;
      }
      if (LoopCount == TimeOut){
        OBU_DEBUG ("OBU: WRITE FLASH SECTOR ERROR\n");
        kfree (OpCode);
        kfree (WriteData);
        return -EFAULT;
      }
      *StatReg = 0xff;
      udelay (1);
    } else {
      while (LoopCount < TimeOut){   // to avoid dead loop
        //
        // Typical Flash Write time is 10us
        //
        udelay (10);

        switch (DataSize){
        case COMMAND_SIZE_BYTE:
          if (*FlashReg8 == ((unsigned char *) WriteData)[Offset]){
            IsWriteComplete = TRUE;
          }
          break;
        case COMMAND_SIZE_WORD:
          if (*FlashReg16 == ((unsigned short *) WriteData)[Offset / DataSize]){
            IsWriteComplete = TRUE;
          }
          break;
        case COMMAND_SIZE_DWORD:
          if (*FlashReg32 == ((unsigned int *) WriteData)[Offset / DataSize]){
            IsWriteComplete = TRUE;
          }
          break;
        }
        if (IsWriteComplete == TRUE){
          break;
        }

        LoopCount++;
      }

      //
      // If timer expired, error return
      //
      if (LoopCount == TimeOut){
        OBU_DEBUG ("OBU: WRITE FLASH SECTOR ERROR\n");
        kfree (OpCode);
        kfree (WriteData);
        return -EFAULT;
      }
    }
  }
  kfree (OpCode);
  kfree (WriteData);
  return STATUS_SUCCESS;
}

/*
 * Erase SPI Flash Sector
 */
static int EraseSpiFlashSector (unsigned long arg, struct _LOCAL_DEVICE_INFO *pLDI)
/*++

Routine Description:
    This routine processes the IOCTLs which erase the Flash Block.

Arguments:
    
    pLDI        - our local device data

Return Value:
    STATUS_SUCCESS           -- OK

    -EFAULT -- Bad address

--*/
{
  unsigned int SectorBase;
  unsigned int SectorSize;
  unsigned int Period;
  unsigned int Flags;
  DWORD SpiControlValue;
  int Status;

  if (get_user (SectorBase, ((unsigned int *) arg)) != 0)
    return -EFAULT;
  if (get_user (SectorSize, ((unsigned int *) arg + 1)) != 0)
    return -EFAULT;
  if (get_user (Flags, ((unsigned int *) arg + 2)) != 0)
    return -EFAULT;
  if (get_user (Period, ((unsigned int *) arg + 3)) != 0)
    return -EFAULT;

  SpiControlValue = pLDI->IchSpi.IchSpiEraseControl;
  Status = IssueSpiCommand (pLDI, SPI_ICH_SET_DATA_SIZE (SpiControlValue, 1), SectorBase, 0, Period * 1000, NULL);
  if (Status != STATUS_SUCCESS){
      return Status;
  }
  return STATUS_SUCCESS;
}

/*
 * Write SPI Flash Sector
 */
static int WriteSpiFlashSector (unsigned long arg, struct _LOCAL_DEVICE_INFO *pLDI)
/*++

Routine Description:
    This routine processes the IOCTLs which write the Flash Block.

Arguments:
    
    pLDI        - our local device data
Return Value:
    STATUS_SUCCESS           -- OK

    -EFAULT -- Bad address

--*/
{
  int Count;
  unsigned long i;

  unsigned long Size;
  unsigned int SectorBase;
  unsigned int SectorSize;
  unsigned int Period;
  unsigned int Flags;
  unsigned char WriteGranularity;
  unsigned char *WriteData;
  unsigned char *Buffer;
  unsigned long SpiFlashAddress;
  DWORD SpiControlValue;
  int Status;
  int isNeedWrite;
  unsigned char *pUbuf = NULL;
  size_t retlen = 0;

  if (get_user (SectorBase, ((unsigned int *) arg)) != 0)
    return -EFAULT;
  if (get_user (SectorSize, ((unsigned int *) arg + 1)) != 0)
    return -EFAULT;
  if (get_user (Flags, ((unsigned int *) arg + 2)) != 0)
    return -EFAULT;
  if (get_user (Period, ((unsigned int *) arg + 3)) != 0)
    return -EFAULT;

  if (SectorSize >= MAX_KMALLOC_SIZE)
    WriteData = (unsigned char *) kmalloc (MAX_KMALLOC_SIZE, GFP_KERNEL);
  else
    WriteData = (unsigned char *) kmalloc (SectorSize, GFP_KERNEL);

  if (!WriteData)
    return -ENOMEM;

  pUbuf = (unsigned char *)((unsigned int *) arg + 4);
  while (SectorSize > 0) {
    if (SectorSize >= MAX_KMALLOC_SIZE)
      Count = MAX_KMALLOC_SIZE;
    else
      Count = SectorSize;

    if (copy_from_user(WriteData, (const void *) (pUbuf + retlen), Count)){
      kfree (WriteData);
      return -EFAULT;
    }
    Buffer = WriteData;
    SpiFlashAddress = SectorBase;

    SpiControlValue = pLDI->IchSpi.IchSpiWriteControl;
    WriteGranularity = pLDI->IchSpi.IchWriteGranularity < SPI_ICH_MAX_DATA_SIZE ? pLDI->IchSpi.IchWriteGranularity : SPI_ICH_MAX_DATA_SIZE;
    while (Count > 0){
      //
      // Specify size to write contiguously.
      //
      Size = (Count - WriteGranularity) > 0 ? WriteGranularity : Count;

      //
      // If write data is 0xff, skip it.
      //
      isNeedWrite = FALSE;
      for (i = 0; i < Size; i++){
        if (Buffer[i] != 0xff){
          isNeedWrite = TRUE;
          break;
        }
      }
      if (isNeedWrite == TRUE){
        Status = IssueSpiCommand (pLDI, SPI_ICH_SET_DATA_SIZE (SpiControlValue, Size), SpiFlashAddress, Size, Period, &Buffer);
        if (Status != STATUS_SUCCESS){
          kfree (WriteData);
          return Status;
        }
      }
      Buffer += Size;
      SpiFlashAddress += Size;
      Count -= Size;
      SectorSize -= Size;
      retlen += Size;
    }
  }

  kfree (WriteData);

  return STATUS_SUCCESS;
}

static int GetSpiJedecId(unsigned long arg, struct _LOCAL_DEVICE_INFO *pLDI)
/*++

Routine Description:
    Read spi id via spi read jedec id command.
Return value:
    STATUS_SUCCESS           -- OK

--*/
{
  int Status;
  unsigned int SpiControlValue;
  unsigned char *Buffer;
  SpiControlValue = pLDI->IchSpi.IchSpiJedecIdControl;
  Buffer = (unsigned char *) kmalloc (sizeof(int), GFP_KERNEL);
  if (copy_from_user(Buffer, (const void *) ((unsigned char *) arg), sizeof(int))){
    kfree (Buffer);
    return -EFAULT;
  }

  Status = IssueSpiCommand(pLDI, SPI_ICH_SET_DATA_SIZE(SpiControlValue, 3), 0, 3, 0, &Buffer);
  Buffer[3] = 0;
#ifdef DEBUG
  printk ("<7>OBU: SPI JEDEC ID:%x %x %x\n", Buffer[0], Buffer[1], Buffer[2]);
#endif

  if (copy_to_user((unsigned char *)arg, Buffer, sizeof(int))){
    kfree (Buffer);
    return -EFAULT;
  }
  return STATUS_SUCCESS;
}

static int IssueSpiCommand (PLOCAL_DEVICE_INFO pLDI, unsigned int ControlValue,
       unsigned int Address, unsigned int Size,
       unsigned int Wait, BYTE ** Buffer)
/*++

Routine Description:
    Issue spi command.
    Assume all registers have alreaady set except address, data, and control.
    Note, this routine does not check any error status.
    Error checking is responsible for the caller.
Return value:
    STATUS_SUCCESS           -- OK

    -EFAULT -- Bad address

--*/
{
  int i;
  unsigned int TimeOut;
  unsigned int LoopCount;
  DWORD CmdType;
  DWORD SpiStatus;

  CmdType = SPI_ICH_GET_CMD_TYPE (ControlValue, pLDI->IchSpi.IchSpiType);
  if (CmdType & SPI_ICH_CMD_ADDRESS_MASK){
    SPI_ICH_WRITE_REG ((unsigned int *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiAddressReg), Address);
  } else {
    SPI_ICH_WRITE_REG((unsigned int *)(pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiAddressReg), 0);
  }

  //
  // If cmd is erase, skip it.
  //
  if(!(CmdType & SPI_ICH_CMD_WRITE_MASK)){
    for (i = 0; i < (int) Size; i++){
      SPI_ICH_WRITE_REG((unsigned char *)pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiDataReg0 + i, 0);
    }
  } else if (CmdType & SPI_ICH_CMD_WRITE_MASK && Buffer != NULL){
    for (i = 0; i < (int) Size; i++){
      SPI_ICH_WRITE_REG ((unsigned char *)pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiDataReg0 + i, *(*Buffer + i));
    }
  }

  SPI_ICH_ALIGNED_WRITE_REG ((unsigned int *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiStatusReg), SPI_ICH_STATUS_DONE_MASK | SPI_ICH_STATUS_ERROR_MASK, pLDI->IchSpi.IchSpiStatusRegSize);

  SPI_ICH_ALIGNED_WRITE_REG ((unsigned int *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiControlReg), ControlValue, pLDI->IchSpi.IchSpiControlRegSize);

  // Waiting until spi controller command complete.
  // Note this loop does not wait device command complate.
  TimeOut = 100;
  LoopCount = 0;
  while (LoopCount < TimeOut){
    if (Wait == 0){
      udelay (1);
    } else {
      udelay (Wait/TimeOut > 1 ? Wait/TimeOut : 1);
    }
    SpiStatus = SPI_ICH_ALIGNED_READ_REG ((unsigned int *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiStatusReg), pLDI->IchSpi.IchSpiStatusRegSize);
    if (!(SpiStatus & SPI_ICH_IN_PROGRESS_MASK) && (SpiStatus & SPI_ICH_STATUS_DONE_MASK)){
      break;
    }
      LoopCount++;
  }
  if (LoopCount == TimeOut){
    return -EFAULT;
  }

  if (!(CmdType & SPI_ICH_CMD_WRITE_MASK)){
    for (i = 0; i < (int) Size; i++){
      *(*Buffer + i) = SPI_ICH_READ_REG ((unsigned char *) (pLDI->pMappedSpiRegs + pLDI->IchSpi.IchSpiDataReg0 + i));
    }
  }

  return STATUS_SUCCESS;
}

/*
 * open device of bios flash rom
 */
static int OBU_open (struct inode *inode, struct file *file)
{
  LOCAL_DEVICE_INFO *pLDI;
  int minor = MINOR (inode->i_rdev);

  if (minor != 0) return -ENODEV;
  spin_lock_irq (&obu_spin);
  if (obu_stat != OBU_CLOSE){
    spin_unlock_irq (&obu_spin);
    return -EBUSY;
  }
  spin_unlock_irq (&obu_spin);

  pLDI = (LOCAL_DEVICE_INFO *) kmalloc (sizeof (LOCAL_DEVICE_INFO), GFP_KERNEL);
  if (!pLDI) return -ENOMEM;

  file->private_data = pLDI;
  pLDI->FlashMapped = FALSE;
  pLDI->SpiUsed = FALSE;
  spin_lock_irq (&obu_spin);
  obu_stat = OBU_OPEN;
  spin_unlock_irq (&obu_spin);

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,0)
  MOD_INC_USE_COUNT;
#endif

  OBU_DEBUG ("OBU: open\n");

  return 0;
}

/*
 * close device of bios flash rom
 */
static int OBU_release (struct inode *inode, struct file *file)
{
  LOCAL_DEVICE_INFO *pLDI = (LOCAL_DEVICE_INFO *) file->private_data;

  OBU_DEBUG ("OBU: close\n");

  spin_lock_irq (&obu_spin);
  if (obu_stat != OBU_CLOSE){
    spin_unlock_irq (&obu_spin);
    if (pLDI->FlashMapped == TRUE){
      iounmap (pLDI->pMappedFlashAddress);
    }
    if (pLDI->SpiUsed == TRUE){
      iounmap (pLDI->pMappedSpiRegs);
    }
    pLDI->FlashMapped = FALSE;
    pLDI->SpiUsed = FALSE;
    spin_lock_irq (&obu_spin);
    obu_stat = OBU_CLOSE;
    spin_unlock_irq (&obu_spin);
    OBU_DEBUG ("OBU: Unmap memory.\n");
  } else {
    spin_unlock_irq (&obu_spin);
  }
  kfree (pLDI);
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,0)
  MOD_DEC_USE_COUNT;
#endif
  return 0;
}

/*
 * This function reads the physical memory mapped bios flash rom.
 */
static ssize_t OBU_read (struct file *file, char *buf, size_t count, loff_t * ppos)
{
  LOCAL_DEVICE_INFO *pLDI = (LOCAL_DEVICE_INFO *) file->private_data;
  unsigned long end = 0;
  unsigned char *buffer = NULL;
  unsigned char *pbuf = NULL;
  ssize_t read = 0;
  ssize_t size = 0;
  size_t retlen = 0;
  size_t out_size = 0;
  int status;
  unsigned int SpiControlValue;

  if (pLDI->FlashMapped){
    end = pLDI->FlashSize;
  } else if (pLDI->SpiUsed == TRUE){
    end = 0xFFFFFFFF;
  } else {
    OBU_DEBUG ("OBU: read: SET FLASH ERROR\n");
    return -EFAULT;
  }

  OBU_DEBUG ("OBU: read\n");

#ifdef DEBUG
  printk ("<7>OBU: position:%x, count:%x, end:%lx\n", (unsigned int)*ppos, (unsigned int)count, end);
#endif
  if (*ppos >= end) return 0;
  if (count > end - *ppos) count = end - *ppos;
  if (!count) return 0;
  if (pLDI->SpiUsed == TRUE){
    if (SPI_ICH_MAX_DATA_SIZE > MAX_KMALLOC_SIZE) return -ENOMEM;
    if (count > MAX_KMALLOC_SIZE)
      buffer = kmalloc(MAX_KMALLOC_SIZE, GFP_KERNEL);
    else
      buffer = kmalloc(count, GFP_KERNEL);

    if (!buffer) return -ENOMEM;

    pbuf = buffer;
    out_size = count;
    SpiControlValue = pLDI->IchSpi.IchSpiReadControl;

    while (out_size > 0){
      if (out_size >= MAX_KMALLOC_SIZE){
        count = MAX_KMALLOC_SIZE;
      } else {
        count = out_size;
      }
      out_size -= count;
      buffer = pbuf;
      retlen = 0;

      while (count > 0){
        size = (count - SPI_ICH_MAX_DATA_SIZE) > 0 ? SPI_ICH_MAX_DATA_SIZE : count;
        status = IssueSpiCommand (pLDI, SPI_ICH_SET_DATA_SIZE (SpiControlValue, size), *ppos, size, 0, &buffer);
        if (status != STATUS_SUCCESS){
          kfree (pbuf);
          return status;
        }
        *ppos += size;
        count -= size;
        read += size;
        buffer += size;
        retlen += size;
      }

      if (copy_to_user (buf, pbuf, retlen)){
        kfree (pbuf);
        return -EFAULT;
      }
      buf += retlen;
    }
#ifdef DEBUG
    printk ("<7>OBU: position:%x, length:%d\n", (int)*ppos, (unsigned int)read);
#endif
    kfree (pbuf);
  } else {
    if (copy_to_user (buf, pLDI->pMappedFlashAddress + *ppos, count))
      return -EFAULT;
    read += count;
    *ppos += read;
  }
    return read;
}

/*
 * seek memory mapped bios flash rom
 */
static loff_t OBU_lseek (struct file *file, loff_t offset, int orig)
{
  LOCAL_DEVICE_INFO *pLDI = (LOCAL_DEVICE_INFO *) file->private_data;
  unsigned long FlashMaxSize;

  if (pLDI->FlashMapped){
    FlashMaxSize = pLDI->FlashSize;
  } else if (pLDI->SpiUsed == TRUE){
    FlashMaxSize = 0xFFFFFFFF;
  } else {
    OBU_DEBUG ("OBU: lseek: SET FLASH ERROR\n");
    return -EFAULT;
  }
  OBU_DEBUG ("OBU: lseek\n");

  switch (orig)
  {
  case 0:
    /* SEEK_SET */
    file->f_pos = offset;
    break;
  case 1:
    /* SEEK_CUR */
    file->f_pos += offset;
  case 2:
    /* SEEK_END */
    file->f_pos = FlashMaxSize + offset;
  default:
    OBU_DEBUG ("OBU: lseek: INVALID OPTION\n");
    return -EINVAL;
  }

#ifdef DEBUG
  printk ("<7>OBU: [lseek] position:%x, offset:%x\n", (unsigned int)file->f_pos, (unsigned int)offset);
#endif
  if (file->f_pos < 0)
    file->f_pos = 0;
  else if (file->f_pos >= FlashMaxSize)
    file->f_pos = FlashMaxSize - 1;

  return file->f_pos;
}

/*
 * ioctl OnlineBIOSUpdate
 */
#if defined(CONFIG_COMPAT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
static long OBU_compat_ioctl (struct file *file, unsigned int cmd, unsigned long arg)
#else
static int OBU_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
#endif
{
  int status = 0;
  unsigned int TmpBuffer;
  UAOpCode OpCode;
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
  int size = _IOC_SIZE (cmd);
#endif
  LOCAL_DEVICE_INFO *pLDI = (LOCAL_DEVICE_INFO *) file->private_data;

  if (_IOC_TYPE (cmd) != OBU_MAGIC) return -EINVAL;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
  if (_IOC_DIR (cmd) & _IOC_READ){
    status = verify_area (VERIFY_WRITE, (void *) arg, size);
  } else if (_IOC_DIR (cmd) & _IOC_WRITE){
      status = verify_area (VERIFY_READ, (void *) arg, size);
  }
#endif
  if (status) return status;
  OBU_DEBUG ("OBU: ioctl verify passed\n");

  switch (cmd)
  {
  case OBU_SET_FLASH_INFO:
    if (pLDI->FlashMapped == TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH INFO IS ALREADY SET\n");
      return -EBUSY;
    }
    OBU_DEBUG ("OBU: ioctl set flash info\n");

    if (get_user (pLDI->FlashBase, ((unsigned int *) arg)) != 0)
      return -EFAULT;
    if (get_user (pLDI->FlashSize, ((unsigned int *) arg) + 1) != 0)
      return -EFAULT;
    if ((pLDI->pMappedFlashAddress =  ioremap_nocache (pLDI->FlashBase, pLDI->FlashSize)) == 0){
      printk (KERN_ERR "OnlineBIOSUpdate: Unable to remap memory.\n");
      return -ENODEV;
    }
#ifdef DEBUG
    printk ("<7>OBU: PhysicalFlashAddr:%x\n", pLDI->FlashBase);
    printk ("<7>OBU: PhysicalFlashSize:%d\n", pLDI->FlashSize);
#endif
    pLDI->FlashMapped = TRUE;
    break;
  case OBU_ERASE_FLASH_SECTOR:
    if (pLDI->FlashMapped != TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH DATA IS NOT SET\n");
      return -EPERM;
    }
    OBU_DEBUG ("OBU: ioctl erase flash sector\n");
    status = EraseFlashSector (arg, pLDI);
    if (status != 0){
      printk ("<3>OnlineBIOSUpdate: ERASE FLASH SECTOR ERROR\n");
      return status;
    }
    break;
  case OBU_WRITE_FLASH_SECTOR:
    if (pLDI->FlashMapped != TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH DATA IS NOT SET\n");
      return -EPERM;
    }
    OBU_DEBUG ("OBU: ioctl write flash sector\n");
    status = WriteFlashSector (arg, pLDI);
    if (status != 0){
      printk ("<3>OnlineBIOSUpdate: WRITE FLASH SECTOR ERROR\n");
      return status;
    }
    break;
  case OBU_SET_SPI_FLASH_INFO:
    if (pLDI->SpiUsed == TRUE){
      iounmap (pLDI->pMappedSpiRegs);
      OBU_DEBUG ("OBU: unmapped spi flash memory\n");
    }
    OBU_DEBUG ("OBU: ioctl set spi flash info\n");
    status = SetSpiFlashInfo (arg, pLDI);
    if (status != 0){
      printk ("<3>OnlineBIOSUpdate: SET SPI FLASH INFO ERROR\n");
      return status;
    }
    break;
  case OBU_ERASE_SPI_FLASH_SECTOR:
    if (pLDI->SpiUsed != TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH DATA IS NOT SET\n");
      return -EPERM;
    }
    OBU_DEBUG ("OBU: ioctl erase spi flash sector\n");
    status = EraseSpiFlashSector (arg, pLDI);
    if (status != 0){
      printk ("<3>OnlineBIOSUpdate: ERASE SPI FLASH SECTOR ERROR\n");
      return status;
    }
    break;
  case OBU_WRITE_SPI_FLASH_SECTOR:
    if (pLDI->SpiUsed != TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH DATA IS NOT SET\n");
      return -EPERM;
    }
    OBU_DEBUG ("OBU: ioctl write flash sector\n");
    status = WriteSpiFlashSector (arg, pLDI);
    if (status != 0){
      printk ("<3>OnlineBIOSUpdate: WRITE SPI FLASH SECTOR ERROR\n");
      return status;
    }
    break;
  case OBU_GET_SPI_FLASH_DATA:
    if (pLDI->SpiUsed != TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH DATA IS NOT SET\n");
      return -EPERM;
    }
    OBU_DEBUG ("OBU: ioctl get spi flash data\n");
    //Can not make use of this ioctl now.
    return -ENOIOCTLCMD;
  case OBU_GET_SPI_JEDEC_ID:
    if (pLDI->SpiUsed != TRUE){
      printk ("<3>OnlineBIOSUpdate: FLASH DATA IS NOT SET\n");
      return -EPERM;
    }
    OBU_DEBUG ("OBU: ioctl get spi jedec id\n");
    status = GetSpiJedecId(arg, pLDI);
    if (status != 0){
      printk ("<3>OnlineBIOSUpdate: GET SPI JEDEC ID ERROR\n");
      return status;
    }
    break;
  case OBU_ACCESS_RESOURCE:
    OBU_DEBUG ("OBU: ioctl access resource\n");

    if (copy_from_user (&OpCode, (const void *) arg, sizeof (OpCode))){
      return -EFAULT;
    }

    OBU_DEBUG ("OBU: cmd=%x, size=%x, addr=%x, data=%x, mask=%x, pad=%x\n",
       OpCode.Cmd, OpCode.CmdSize, OpCode.Addr, OpCode.Data,
       OpCode.Mask, OpCode.Pad[0]);

    switch (CHECK_COMMAND_TYPE (OpCode.Cmd))
    {
    case COMMAND_TYPE_READ:
      switch (CHECK_RESOURCE_TYPE (OpCode.Cmd))
      {
      case RESOURCE_TYPE_IO:
        status = ReadIo (OpCode, &pLDI->Register0);
        return status;
      case RESOURCE_TYPE_MEMORY:
        status = ReadMemory (OpCode, &pLDI->Register0);
        return status;
      case RESOURCE_TYPE_PCI:
        status = ReadPci (OpCode, &pLDI->Register0);
        return status;
      default:
        printk ("<3>OnlineBIOSUpdate: RESOURCE TYPE READ ERROR\n");
        return -EFAULT;
      }
    case COMMAND_TYPE_WRITE:
      switch (CHECK_RESOURCE_TYPE (OpCode.Cmd))
      {
      case RESOURCE_TYPE_IO:
        return WriteIo (OpCode);
      case RESOURCE_TYPE_MEMORY:
        return WriteMemory (OpCode);
      case RESOURCE_TYPE_PCI:
        return WritePci (OpCode);
      default:
        printk ("<3>OnlineBIOSUpdate: RESOURCE TYPE WRITE ERROR\n");
        return -EFAULT;
      }
    case COMMAND_TYPE_SET:
    case COMMAND_TYPE_RESET:
      if (OpCode.Pad[0] == 1){
        TmpBuffer = pLDI->Register0;
      } else {
        switch (CHECK_RESOURCE_TYPE (OpCode.Cmd))
        {
        case RESOURCE_TYPE_IO:
          status = ReadIo (OpCode, &TmpBuffer);
          break;
        case RESOURCE_TYPE_MEMORY:
          status = ReadMemory (OpCode, &TmpBuffer);
          break;
        case RESOURCE_TYPE_PCI:
          status = ReadPci (OpCode, &TmpBuffer);
          break;
        default:
          printk("<3>OnlineBIOSUpdate: RESOURCE TYPE SET/RESET ERROR\n");
        return -EFAULT;
        }
      }
      if (status != 0){
        printk("<3>OnlineBIOSUpdate: RESOURCE TYPE SET/RESET READ ERROR\n");
        return -EFAULT;
      }
      if (CHECK_COMMAND_TYPE (OpCode.Cmd) == COMMAND_TYPE_SET){
        TmpBuffer |= OpCode.Mask;
      } else {
        TmpBuffer &= OpCode.Mask;
      }
      OpCode.Data = TmpBuffer;

      switch (CHECK_RESOURCE_TYPE (OpCode.Cmd))
      {
      case RESOURCE_TYPE_IO:
        return WriteIo (OpCode);
      case RESOURCE_TYPE_MEMORY:
        return WriteMemory (OpCode);
      case RESOURCE_TYPE_PCI:
        return WritePci (OpCode);
      default:
        printk ("<3>OnlineBIOSUpdate: RESOURCE TYPE ERROR\n");
        return -EFAULT;
      }
    default:
      printk ("<3>OnlineBIOSUpdate: COMMAND TYPE ERROR\n");
      return -EFAULT;
    }
    break;
  case OBU_GET_MEMORY_DATA:
    OBU_DEBUG ("OBU: ioctl get memory data\n");
    //Can not make use of this ioctl now.
    return -ENOIOCTLCMD;
  default:
    printk ("<3>OnlineBIOSUpdate: IOCTL COMMAND ERROR\n");
    return -ENOIOCTLCMD;
  }
  return status;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static struct file_operations OBU_fops = {
  owner   :THIS_MODULE,
  llseek  :OBU_lseek,
  read    :OBU_read,
  ioctl   :OBU_ioctl,
  open    :OBU_open,
  release :OBU_release,
};
#else
static struct file_operations OBU_fops = {
  .owner   = THIS_MODULE,
  .llseek  = OBU_lseek,
  .read    = OBU_read,
#if defined(CONFIG_COMPAT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
  .unlocked_ioctl = OBU_compat_ioctl,
#else
  .ioctl   = OBU_ioctl,
#endif
#if defined(CONFIG_COMPAT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
  .compat_ioctl = OBU_compat_ioctl,
#endif
  .open    = OBU_open,
  .release = OBU_release,
};
#endif

static int __init init_OBU (void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  struct class_device *class_err;
#endif
  int rv = 0;

  rv = register_chrdev (obu_major, "OnlineBIOSUpdate", &OBU_fops);
  if (rv < 0){
    printk (KERN_ERR "OnlineBIOSUpdate: Can't allocate major number\n");
    goto out;
  }
  obu_major = rv;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  obu_class = class_create (THIS_MODULE, "OnlineBIOSUpdate");
  if (IS_ERR (obu_class)){
    printk (KERN_ERR "OnlineBIOSUpdate: Can't register device class\n");
    rv = PTR_ERR (obu_class);
    goto out_chrdev;
  }

  class_err = class_device_create (obu_class, NULL, MKDEV (obu_major, 0), NULL, "OnlineBIOSUpdate");
  if (IS_ERR (class_err)){
    rv = PTR_ERR (class_err);
    goto out_class;
  }
#endif

#if defined(__x86_64__) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
  rv = register_ioctl32_conversion (OBU_SET_FLASH_INFO,           NULL);
  rv |= register_ioctl32_conversion (OBU_ERASE_FLASH_SECTOR,      NULL);
  rv |= register_ioctl32_conversion (OBU_WRITE_FLASH_SECTOR,      NULL);
  rv |= register_ioctl32_conversion (OBU_ACCESS_RESOURCE,         NULL);
  rv |= register_ioctl32_conversion (OBU_GET_MEMORY_DATA,        NULL);
  rv |= register_ioctl32_conversion (OBU_SET_SPI_FLASH_INFO,     NULL);
  rv |= register_ioctl32_conversion (OBU_ERASE_SPI_FLASH_SECTOR, NULL);
  rv |= register_ioctl32_conversion (OBU_WRITE_SPI_FLASH_SECTOR, NULL);
  rv |= register_ioctl32_conversion (OBU_GET_SPI_FLASH_DATA,     NULL);
  rv |= register_ioctl32_conversion (OBU_GET_SPI_JEDEC_ID,       NULL);
  if (rv){
    unregister_ioctl32_conversion (OBU_SET_FLASH_INFO);
    unregister_ioctl32_conversion (OBU_ERASE_FLASH_SECTOR);
    unregister_ioctl32_conversion (OBU_WRITE_FLASH_SECTOR);
    unregister_ioctl32_conversion (OBU_ACCESS_RESOURCE);
    unregister_ioctl32_conversion (OBU_GET_MEMORY_DATA);
    unregister_ioctl32_conversion (OBU_SET_SPI_FLASH_INFO);
    unregister_ioctl32_conversion (OBU_ERASE_SPI_FLASH_SECTOR);
    unregister_ioctl32_conversion (OBU_WRITE_SPI_FLASH_SECTOR);
    unregister_ioctl32_conversion (OBU_GET_SPI_FLASH_DATA);
    unregister_ioctl32_conversion (OBU_GET_SPI_JEDEC_ID);
    rv = -EBUSY;
    goto out_chrdev;
  }
#endif

  OBU_DEBUG ("OBU: driver installed: major = %d\n", obu_major);

  rv = 0;
  goto out;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
out_class:
  class_device_destroy (obu_class, MKDEV (obu_major, 0));
  class_destroy (obu_class);
#endif

#if defined(__x86_64__) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
out_chrdev:
  unregister_chrdev (obu_major, "OnlineBIOSUpdate");
#endif

out:
  return rv;
}

static void __exit cleanup_OBU (void)
{
  OBU_DEBUG ("OBU: driver uninstalled\n");

#if defined(__x86_64__) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
  unregister_ioctl32_conversion (OBU_SET_FLASH_INFO);
  unregister_ioctl32_conversion (OBU_ERASE_FLASH_SECTOR);
  unregister_ioctl32_conversion (OBU_WRITE_FLASH_SECTOR);
  unregister_ioctl32_conversion (OBU_ACCESS_RESOURCE);
  unregister_ioctl32_conversion (OBU_GET_MEMORY_DATA);
  unregister_ioctl32_conversion (OBU_SET_SPI_FLASH_INFO);
  unregister_ioctl32_conversion (OBU_ERASE_SPI_FLASH_SECTOR);
  unregister_ioctl32_conversion (OBU_WRITE_SPI_FLASH_SECTOR);
  unregister_ioctl32_conversion (OBU_GET_SPI_FLASH_DATA);
  unregister_ioctl32_conversion (OBU_GET_SPI_JEDEC_ID);
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  class_device_destroy (obu_class, MKDEV (obu_major, 0));
  class_destroy (obu_class);
#endif

  unregister_chrdev (obu_major, "OnlineBIOSUpdate");
}

module_init (init_OBU);
module_exit (cleanup_OBU);
